@@ -4,7 +4,7 @@ use 5.010;
 use strict;
 use warnings;
 
-our $VERSION = '0.14';
+our $VERSION = '0.15';
 
 use Carp;
 use Scalar::Util qw( set_prototype );
@@ -67,6 +67,12 @@ sub import {
     *{"${caller}::$shortcut"}     = $silo;
 };
 
+sub get_meta {
+    my ($self, $target) = @_;
+
+    return $metadata{$target};
+}
+
 1; # End of Resource::Silo
 
 __END__
@@ -323,13 +329,17 @@ itself except that it's more explicit.
 
 =head3 dependencies => \@list
 
-List other resources that may be requested in the initializer.
-Unless C<loose_deps> is specified (see below),
-the dependencies I<must> be declared I<before> the dependant.
+List of other resources that may be requested in the initializer.
+Attempting to use a resource that was not listed will cause an exception.
 
-A resource with parameter may also depend on itself.
+Attempt to create a circular dependency will cause an exception as well.
+As an exception, a resource with parameter may depend on itself.
 
-The default is all eligible resources known so far.
+Lastly, having unresolved dependencies upon container instantiation
+will also cause an exception.
+
+If unspecified, any dependencies are allowed, but circular instantiation
+will still be prohibited.
 
 B<NOTE> This behavior was different prior to v.0.09
 and may be change again in the near future.
@@ -337,16 +347,6 @@ and may be change again in the near future.
 This parameter has a different structure
 if C<class> parameter is in action (see below).
 
-=head3 loose_deps => 0 | 1
-
-Allow dependencies that have not been declared yet.
-
-Not specifying the C<dependencies> parameter would now mean
-there are no restrictions whatsoever.
-
-B<NOTE> Having to resort to this flag may be
-a sign of a deeper architectural problem.
-
 =head3 argument => C<sub { ... }> || C<qr( ... )>
 
 Declare a (possibly infinite) set of sibling resources under the same name,
@@ -379,6 +379,16 @@ E.g. when using L<Redis::Namespace>:
             );
         };
 
+=head3 check => sub { $container, $resource, $name, $argument }
+
+Check the resource after initialization.
+The function must throw an exception if the resource is invalid.
+
+It is applied after the resource is initialized and before it is put into the cache,
+both for normally initialized and overridden resources.
+
+B<NOTE> Experimental, name and semantics may change in the future.
+
 =head3 cleanup => sub { $resource_instance }
 
 Undo the init procedure.
@@ -452,6 +462,24 @@ not during request processing.
 
 See L<Resource::Silo::Container/preload>.
 
+=head3 nullable => 0 | 1
+
+Setting this flag allows returning C<undef> from the initializer
+without causing an exception.
+
+The undefined value will still be cached, so the initializer
+will not be called again for the same resource (and argument, if any).
+
+This may indicate e.g. that the resource is optional
+and is not required for normal operations.
+
+IF C<check> is present, it will be omitted for C<undef> value.
+
+=head3 loose_deps => 0 | 1
+
+Prior to v.0.14, this flag allowed forward dependencies.
+Currently does nothing except emitting a warning.
+
 =head2 silo
 
 A re-exportable function returning one and true container instance
@@ -468,6 +496,13 @@ C<silo-E<gt>new> will create a new instance of the I<same> container class.
 The resource container class may therefore be viewed as an
 I<optional singleton>.
 
+=head1 UTILITY FUNCTIONS
+
+=head2 get_meta($package)
+
+Return a L<Resource::Silo::Metadata> object associated with the given package,
+if any.
+
 =head1 CAVEATS AND CONSIDERATIONS
 
 See L<Resource::Silo::Container> for the container implementation.
