@@ -2,7 +2,7 @@ package Resource::Silo::Container;
 
 use strict;
 use warnings;
-our $VERSION = '0.14';
+our $VERSION = '0.15';
 
 =head1 NAME
 
@@ -51,6 +51,10 @@ sub BUILD {
     my $spec = $Resource::Silo::metadata{ref $self}
         // _silo_find_metaclass($self);
 
+    # TODO Carp doesn't seem to work correctly from constructor with either Moo or Moose,
+    #      but at least we now generate pointers to where the offending resources are declared
+    $spec->run_pending_checks;
+
     $self->{-spec} = $spec;
     $self->{-pid} = $$;
 
@@ -149,8 +153,15 @@ sub _silo_instantiate_res {
     local $self->{-pending}{$key} = 1;
     local $self->{-allow} = $spec->{allowdeps};
 
-    ($self->{-override}{$name} // $spec->{init})->($self, $name, $arg)
-        // croak "Instantiating resource '$key' $spec->{origin} failed for no apparent reason";
+    my $init = $self->{-override}{$name} // $spec->{init};
+    my $entity = $init->($self, $name, $arg);
+    if (!defined $entity) {
+        return $entity if ($spec->{nullable});
+        croak "Instantiating resource '$key' $spec->{origin} returned undef for no apparent reason";
+    }
+    $spec->{check}->($self, $entity, $name, $arg)
+        if $spec->{check};
+    return $entity;
 };
 
 # use instead of delete $self->{-cache}{$name}
@@ -213,7 +224,10 @@ sub _silo_make_accessor {
         # Stringify $arg ASAP, we'll validate it inside _silo_instantiate_res().
         # The cache entry for an invalid argument will never get populated.
         my $key = defined $arg && !ref $arg ? $arg : '';
-        $self->{-cache}{$name}{$key} //= $self->_silo_instantiate_res($name, $arg);
+        if (!exists $self->{-cache}{$name}{$key}) {
+            $self->{-cache}{$name}{$key} = $self->_silo_instantiate_res($name, $arg);
+        };
+        return $self->{-cache}{$name}{$key};
     };
 };
 
@@ -374,7 +388,7 @@ sub preload {
 
     my $meta = $$self->{-spec};
 
-    $meta->self_check;
+    $meta->preload;
 
     my $list = $meta->{preload};
     for my $name (@$list) {
